using System;
using System.Collections.Generic;
using System.Threading;
using Dashboard.Windowing;

namespace Dashboard
{
    public class Application : IDisposable
    {
        public IDashboardBackend Backend { get; }
        public IWindowFactory WindowFactory { get; set; }

        public string Name { get; } = "Dashboard Application";

        protected bool IsInitialized { get; private set; } = false;
        protected bool IsDisposed { get; private set; } = false;

        private readonly List<IPhysicalWindow> _physicalWindows = new List<IPhysicalWindow>();

        public event EventHandler? PreInitializing;
        public event EventHandler? Initializing;
        public event EventHandler? PostInitializing;
        public event EventHandler? Leaving;

        public Application(IDashboardBackend backend)
        {
            Backend = backend;
            WindowFactory = backend;
        }

        protected virtual void PreInitialize()
        {
            PreInitializing?.Invoke(this, EventArgs.Empty);
        }

        public virtual void Initialize()
        {
            Backend.Initialize();
            Initializing?.Invoke(this, EventArgs.Empty);
        }

        protected virtual void PostInitialize()
        {
            PostInitializing?.Invoke(this, EventArgs.Empty);
        }

        private void InitializeInternal()
        {
            if (IsInitialized)
                return;

            IsInitialized = true;

            PreInitialize();
            Initialize();
            PostInitialize();
        }

        public virtual void RunEvents(bool wait)
        {
            if (!IsInitialized)
                throw new InvalidOperationException("The application is not initialized. Cannot run events at this time.");

            Backend.RunEvents(wait);
        }

        public virtual void Leave()
        {
            Backend.Leave();
        }

        public void Run() => Run(true, CancellationToken.None);
        public void Run(bool wait) => Run(wait, CancellationToken.None);
        
        public void Run(bool waitForEvents, CancellationToken token)
        {
            InitializeInternal();

            while (!token.IsCancellationRequested)
            {
                RunEvents(waitForEvents);

                foreach (IPhysicalWindow window in _physicalWindows)
                {
                    window.Paint();
                }
            }

            Leave();
        }

        /// <inheritdoc cref="IWindowFactory.CreateWindow"/>
        public IWindow CreateWindow()
        {
            IWindow window = WindowFactory.CreateWindow();

            if (window is IPhysicalWindow physical)
                _physicalWindows.Add(physical);

            return window;
        }

        /// <inheritdoc cref="IWindowFactory.CreatePhysicalWindow"/>
        public IPhysicalWindow CreatePhysicalWindow()
        {
            IPhysicalWindow window = WindowFactory.CreatePhysicalWindow();
            _physicalWindows.Add(window);
            return window;
        } 

        protected virtual void Dispose(bool disposing)
        {
        }

        protected void InvokeDispose(bool disposing)
        {
            if (IsDisposed)
                return;

            IsDisposed = true;

            Dispose(disposing);

            if (disposing)
                GC.SuppressFinalize(this);
        }

        public void Dispose() => InvokeDispose(true);
    }
}